from django.test import TestCase
from casexml.apps.phone.models import SyncLog, User
from couchforms.models import XFormInstance
from casexml.apps.case.models import CommCareCase
from datetime import datetime
from casexml.apps.phone.restore import generate_restore_payload,\
    generate_restore_response, BadStateException
from casexml.apps.phone.checksum import EMPTY_HASH, CaseStateHash
from casexml.apps.case.xml import V2
from casexml.apps.case.util import post_case_blocks
from casexml.apps.case.tests.util import CaseBlock, delete_all_sync_logs, delete_all_xforms, delete_all_cases


class StateHashTest(TestCase):
    
    def setUp(self):
        delete_all_cases()
        delete_all_xforms()
        delete_all_sync_logs()

        self.user = User(user_id="state_hash", username="state_hash",
                         password="changeme", date_joined=datetime(2011, 6, 9)) 
        
        # this creates the initial blank sync token in the database
        generate_restore_payload(self.user)
        [sync_log] = SyncLog.view("phone/sync_logs_by_user", include_docs=True, reduce=False).all()
        self.sync_log = sync_log
    
    def testEmpty(self):
        empty_hash = CaseStateHash(EMPTY_HASH)
        wrong_hash = CaseStateHash("thisisntright")
        self.assertEqual(empty_hash, self.sync_log.get_state_hash())
        response = generate_restore_response(self.user, self.sync_log.get_id, version=V2)
        self.assertEqual(200, response.status_code)
        
        try:
            response = generate_restore_payload(self.user, self.sync_log.get_id,
                                                version=V2, state_hash=str(wrong_hash))
            self.fail("Call to generate a payload with a bad hash should fail!")
        except BadStateException, e:
            self.assertEqual(empty_hash, e.expected)
            self.assertEqual(wrong_hash, e.actual)
            self.assertEqual(0, len(e.case_ids))

        response = generate_restore_response(self.user, self.sync_log.get_id, version=V2,
                                             state_hash=str(wrong_hash))
        self.assertEqual(412, response.status_code)
        
    def testMismatch(self):
        self.assertEqual(CaseStateHash(EMPTY_HASH), self.sync_log.get_state_hash())
        
        c1 = CaseBlock(case_id="abc123", create=True, 
                       owner_id=self.user.user_id).as_xml()
        c2 = CaseBlock(case_id="123abc", create=True, 
                       owner_id=self.user.user_id).as_xml()
        post_case_blocks([c1, c2], 
                         form_extras={"last_sync_token": self.sync_log.get_id})
        
        self.sync_log = SyncLog.get(self.sync_log.get_id)
        real_hash = CaseStateHash("409c5c597fa2c2a693b769f0d2ad432b")
        bad_hash = CaseStateHash("thisisntright")
        self.assertEqual(real_hash, self.sync_log.get_state_hash())
        generate_restore_payload(self.user, self.sync_log.get_id,
                                 version=V2, state_hash=str(real_hash))
        
        try:
            generate_restore_payload(self.user, self.sync_log.get_id,
                                                version=V2, state_hash=str(bad_hash))
            self.fail("Call to generate a payload with a bad hash should fail!")
        except BadStateException, e:
            self.assertEqual(real_hash, e.expected)
            self.assertEqual(bad_hash, e.actual)
            self.assertEqual(2, len(e.case_ids))
            self.assertTrue("abc123" in e.case_ids)
            self.assertTrue("123abc" in e.case_ids)
